/*
 *  linux/arch/arm/mach-pxa/generic.c
 *
 *  Author:	Nicolas Pitre
 *  Created:	Jun 15, 2001
 *  Copyright:	MontaVista Software Inc.
 * 
 * Code common to all PXA machines.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * Since this file should be linked before any other machine specific file,
 * the __initcall() here will be executed first.  This serves as default
 * initialization stuff for PXA machines which can be overriden later if
 * need be.
 */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/pm.h>

#include <asm/hardware.h>
#include <asm/system.h>
#include <asm/pgtable.h>
#include <asm/mach/map.h>

#ifdef CONFIG_PXA_RTC_HACK
#include <asm/setup.h>
#include <linux/bootmem.h>
#endif

#include "generic.h"

/*
 * Various clock factors driven by the CCCR register.
 */

/* Crystal Frequency to Memory Frequency Multiplier (L) */
static unsigned char L_clk_mult[32] = { 0, 27, 32, 36, 40, 45, 0, };

/* Memory Frequency to Run Mode Frequency Multiplier (M) */
static unsigned char M_clk_mult[4] = { 0, 1, 2, 4 };

/* Run Mode Frequency to Turbo Mode Frequency Multiplier (N) */
/* Note: we store the value N * 2 here. */
static unsigned char N2_clk_mult[8] = { 0, 0, 2, 3, 4, 0, 6, 0 };

/* Crystal clock */
#define BASE_CLK	3686400

/*
 * Get the clock frequency as reflected by CCCR and the turbo flag.
 * We assume these values have been applied via a fcs.
 * If info is not 0 we also display the current settings.
 */
unsigned int get_clk_frequency_khz( int info)
{
	unsigned long cccr, turbo;
	unsigned int l, L, m, M, n2, N;

	cccr = CCCR;
	asm( "mrc\tp14, 0, %0, c6, c0, 0" : "=r" (turbo) );

	l  =  L_clk_mult[(cccr >> 0) & 0x1f];
	m  =  M_clk_mult[(cccr >> 5) & 0x03];
	n2 = N2_clk_mult[(cccr >> 7) & 0x07];

	L = l * BASE_CLK;
	M = m * L;
	N = n2 * M / 2;

	if( info)
	{
		L += 5000;
		printk( KERN_INFO "Memory clock: %d.%02dMHz (*%d)\n",
			L / 1000000, (L % 1000000) / 10000, l );
		M += 5000;
		printk( KERN_INFO "Run Mode clock: %d.%02dMHz (*%d)\n",
			M / 1000000, (M % 1000000) / 10000, m );
		N += 5000;
		printk( KERN_INFO "Turbo Mode clock: %d.%02dMHz (*%d.%d, %sactive)\n",
			N / 1000000, (N % 1000000) / 10000, n2 / 2, (n2 % 2) * 5,
			(turbo & 1) ? "" : "in" );
	}

	return (turbo & 1) ? (N/1000) : (M/1000);
}

EXPORT_SYMBOL(get_clk_frequency_khz);

/*
 * Return the current lclk requency in units of 10kHz
 */
unsigned int get_lclk_frequency_10khz(void)
{
	return L_clk_mult[(CCCR >> 0) & 0x1f] * BASE_CLK / 10000;
}

EXPORT_SYMBOL(get_lclk_frequency_10khz);

/*
 * Handy function to set GPIO alternate functions
 */

void set_GPIO_mode(int gpio_mode)
{
	long flags;
	int gpio = gpio_mode & GPIO_MD_MASK_NR;
	int dir = gpio_mode & GPIO_MD_MASK_DIR;
	int fn = (gpio_mode & GPIO_MD_MASK_FN) >> 8;
	int gafr;

	if (0) printk("set_GPIO_mode: gpio=%d dir=%d fn=%d\n", gpio, dir, fn);
	local_irq_save(flags);
	if (gpio_mode & GPIO_MD_MASK_DIR) {
		/* if output and active low, then first set the bit to make it inactive */
		if (gpio_mode & GPIO_ACTIVE_LOW)
			GPSR(gpio) = GPIO_bit(gpio);
		else /* otherwise clear the bit */
			GPCR(gpio) = GPIO_bit(gpio);
		GPDR(gpio) |= GPIO_bit(gpio);
	} else {
		GPDR(gpio) &= ~GPIO_bit(gpio);
	}
	gafr = GAFR(gpio) & ~(0x3 << (((gpio) & 0xf)*2));
	GAFR(gpio) = gafr |  (fn  << (((gpio) & 0xf)*2));
	if (0) printk("\t GPDR(%d)=0x%08x GAFR(%d)=0x%08x\n", (gpio >> 5), GPDR(gpio), (gpio >> 4), GAFR(gpio));
	local_irq_restore(flags);
}

EXPORT_SYMBOL(set_GPIO_mode);

/* 
 * Note that 0xfffe0000-0xffffffff is reserved for the vector table and
 * cache flush area.
 */
static struct map_desc standard_io_desc[] __initdata = {
 /* virtual     physical    length      domain     r  w  c  b */
  { 0xf6000000, 0x20000000, 0x01000000, DOMAIN_IO, 0, 1, 0, 0 }, /* PCMCIA0 IO */
  { 0xf7000000, 0x30000000, 0x01000000, DOMAIN_IO, 0, 1, 0, 0 }, /* PCMCIA1 IO */
  { 0xf8000000, 0x40000000, 0x01800000, DOMAIN_IO, 0, 1, 0, 0 }, /* Devs */
  { 0xfa000000, 0x44000000, 0x00100000, DOMAIN_IO, 0, 1, 0, 0 }, /* LCD */
  { 0xfc000000, 0x48000000, 0x00100000, DOMAIN_IO, 0, 1, 0, 0 }, /* Mem Ctl */
  { 0xff000000, 0x00000000, 0x00100000, DOMAIN_IO, 0, 1, 0, 0 }, /* UNCACHED_PHYS_0 */
  LAST_DESC
};

void __init pxa_map_io(void)
{
	iotable_init(standard_io_desc);
	get_clk_frequency_khz( 1);
#ifdef CONFIG_PXA_RTC_HACK
	pxa_rtc_hack_init();
#endif

#if defined CONFIG_SERIAL_SIR_PXA
        /* Enable IrDA UART (SIR)*/
        CKEN |= CKEN5_STUART;

        /* We want to get our goods from the STUART */
        set_GPIO_mode(GPIO46_STRXD_MD);
        set_GPIO_mode(GPIO47_STTXD_MD);

        /* make sure FIR ICP is off */
        ICCR0 = 0;

        /* configure STUART to for SIR
         * NOTE: RCVEIR and XMITIR must not be set at the same time!
         * Start with receive in IR mode, and switch transmit to IR only
         * when we need to send something in serial driver.
         */
        STISR = STISR_RXPL | STISR_XMODE | STISR_RCVEIR/* | STISR_XMITIR*/;
#endif
}



#ifdef CONFIG_PXA_RTC_HACK
unsigned long *save_RCNR = 0;

void pxa_rtc_hack_init(void)
{
	/* 
	   This has to be here since I guess the bootmem API is the
	   right choice to allocate the memory during boot
	   place. And we are sure that timer iqr is not already
	   running.
	   - Christian Pellegin <chri@infis.univ.trieste.it>
	*/
	unsigned long pxa_rtc_hack = 0;

	pxa_rtc_hack = meminfo.bank[meminfo.nr_banks-1].start + 
		meminfo.bank[meminfo.nr_banks-1].size -
		PAGE_SIZE;
	reserve_bootmem(pxa_rtc_hack, PAGE_SIZE);
	printk("Reserved %ld bytes at %lx for RTC hack\n", 
	       PAGE_SIZE, pxa_rtc_hack);
	save_RCNR = (unsigned long *)  phys_to_virt(pxa_rtc_hack);
	if ( (save_RCNR[0] ^ save_RCNR[1]) == 0xffffffff ) {
		printk("Restoring saved RCNR value to %ld (from %lx)\n", 
		       save_RCNR[0], (unsigned long) save_RCNR);
		RCNR = save_RCNR[0];
	}
	else {
		printk("No valid saved RCNR value found at %lx\n", (unsigned long) save_RCNR);
	}
}
#endif /* CONFIG_PXA_RTC_HACK */
